@@ -6,6 +6,7 @@ from collections import defaultdict
 
 from .elf import ELF
 from ..blob import Blob
+from ..region import Segment
 from .. import register_backend
 from ...errors import CLEError, CLECompatibilityError
 from ...memory import Clemory
@@ -19,17 +20,45 @@ l = logging.getLogger(name=__name__)
 class ELFCore(ELF):
     """
     Loader class for ELF core files.
+
+    One key pain point when analyzing a core dump generated on a remote machine is that the paths to binaries are
+    absolute (and may not exist or be the same on your local machine).
+
+    Therefore, you can use the options ```remote_file_mapping`` to specify a ``dict`` mapping (easy if there are a small
+    number of mappings) or ``remote_file_mapper`` to specify a function that accepts a remote file name and returns the
+    local file name (useful if there are many mappings).
+
+    If you specify both ``remote_file_mapping`` and ``remote_file_mapper``, ``remote_file_mapping`` is applied first,
+    then the result is passed to ``remote_file_mapper``.
+
+    :param executable:           Optional path to the main binary of the core dump. If not supplied, ELFCore will
+                                 attempt to figure it out automatically from the core dump.
+    :param remote_file_mapping:  Optional dict that maps specific file names in the core dump to other file names.
+    :param remote_file_mapper:   Optional function that is used to map every file name in the core dump to whatever is
+                                 returned from this function.
     """
     is_default = True # Tell CLE to automatically consider using the ELFCore backend
 
-    def __init__(self, *args, executable=None, **kwargs):
+    def __init__(self, *args, executable=None, remote_file_mapping=None, remote_file_mapper=None, **kwargs):
         super().__init__(*args, **kwargs)
 
         self.filename_lookup = []
         self.__current_thread = None
         self._threads = []
         self.auxv = {}
+        self.pr_fname = None
         self._main_filepath = executable
+        self._page_size = 0x1000 # a default page size, will be changed later by parsing notes
+        self._main_object = None
+
+        if remote_file_mapping is not None:
+            self._remote_file_mapper = lambda x: remote_file_mapping.get(x, x)
+        else:
+            self._remote_file_mapper = lambda x: x
+
+        if remote_file_mapper is not None:
+            orig = self._remote_file_mapper
+            self._remote_file_mapper = lambda x: remote_file_mapper(orig(x))
 
         self.__extract_note_info()
 
@@ -69,13 +98,20 @@ class ELFCore(ELF):
                 for note in seg_readelf.iter_notes():
                     if note.n_type == 'NT_PRSTATUS':
                         self.__cycle_thread()
-                        self.__parse_prstatus(note.n_desc.encode('latin-1'))  # ???
+                        n_desc = note.n_desc.encode('latin-1') if isinstance(note.n_desc, str) else note.n_desc
+                        self.__parse_prstatus(n_desc)
+                    elif note.n_type == 'NT_PRPSINFO':
+                        self.__parse_prpsinfo(note.n_desc)
+                    elif note.n_type == 'NT_AUXV':
+                        n_desc = note.n_desc.encode('latin-1') if isinstance(note.n_desc, str) else note.n_desc
+                        self.__parse_auxv(n_desc)
                     elif note.n_type == 'NT_FILE':
                         self.__parse_files(note.n_desc)
-                    elif note.n_type == 'NT_AUXV':
-                        self.__parse_auxv(note.n_desc.encode('latin-1'))
                     elif note.n_type == 512 and self.arch.name == 'X86':
-                        self.__parse_x86_tls(note.n_desc.encode('latin-1'))
+                        n_desc = note.n_desc.encode('latin-1') if isinstance(note.n_desc, str) else note.n_desc
+                        self.__parse_x86_tls(n_desc)
+
+        self._replace_main_object_path()
 
         self.__cycle_thread()
         if not self._threads:
@@ -93,6 +129,27 @@ class ELFCore(ELF):
                 for thread in self._threads:
                     thread['segments'] = {thread['registers']['gs'] >> 3: (0, 0xffffffff, 0x51)}
 
+    def _replace_main_object_path(self):
+        """
+        try to replace path of the main_object with the specified one
+        """
+        if not self._main_filepath or not self.filename_lookup:
+            return
+
+        # identify the original path and assuming pr_fname always exists
+        matched = None
+        for i, (a, b, c, fn) in enumerate(self.filename_lookup):
+            if os.path.basename(fn).startswith(self.pr_fname): # pr_fname is defined to be the first 16 bytes of the executable name
+                matched = fn
+                break
+        else:
+            raise CLEError("Fail to find the main object, is this core dump malformed?")
+
+        # replace the path
+        for i, (a, b, c, fn) in enumerate(self.filename_lookup):
+            if fn == matched:
+                self.filename_lookup[i] = (a, b, c, self._main_filepath)
+
     @property
     def __dummy_clemory(self):
         dummy_clemory = Clemory(self.arch, root=True)
@@ -102,20 +159,13 @@ class ELFCore(ELF):
 
     def __parse_prstatus(self, desc):
         """
-        Parse out the prstatus, accumulating the general purpose register values. Supports AMD64, X86, ARM, and AARCH64
-        at the moment.
+        Parse out the prstatus, accumulating the general purpose register values.
+        Supports AMD64, X86, ARM, AArch64, MIPS and MIPSEL at the moment.
 
         :param prstatus: a note object of type NT_PRSTATUS.
         """
 
         # TODO: support all architectures angr supports
-
-        result = {}
-        result['si_signo'], result['si_code'], result['si_errno'] = struct.unpack("<3I", desc[:12])
-
-        # this field is a short, but it's padded to an int
-        result['pr_cursig'] = struct.unpack("<I", desc[12:16])[0]
-
         arch_bytes = self.arch.bytes
         if arch_bytes == 4:
             fmt = "I"
@@ -124,29 +174,38 @@ class ELFCore(ELF):
         else:
             raise CLEError("Architecture must have a bitwidth of either 64 or 32")
 
-        result['pr_sigpend'], result['pr_sighold'] = struct.unpack("<" + (fmt * 2), desc[16:16+(2*arch_bytes)])
+        end = '>' if self.arch.memory_endness == 'Iend_BE' else '<'
 
-        attrs = struct.unpack("<IIII", desc[16+(2*arch_bytes):16+(2*arch_bytes)+(4*4)])
-        result['pr_pid'], result['pr_ppid'], result['pr_pgrp'], result['pr_sid'] = attrs
+        pos = 0
 
-        # parse out the 4 timevals
-        pos = 16+(2*arch_bytes)+(4*4)
-        usec = struct.unpack("<" + fmt, desc[pos:pos+arch_bytes])[0] * 1000
-        result['pr_utime_usec'] = struct.unpack("<" + fmt, desc[pos+arch_bytes:pos+arch_bytes*2])[0] + usec
+        def read_longs(n):
+            fin = pos+n*arch_bytes
+            return (fin, *struct.unpack(end + fmt * n, desc[pos:fin]))
 
-        pos += arch_bytes * 2
-        usec = struct.unpack("<" + fmt, desc[pos:pos+arch_bytes])[0] * 1000
-        result['pr_stime_usec'] = struct.unpack("<" + fmt, desc[pos+arch_bytes:pos+arch_bytes*2])[0] + usec
+        def read_ints(n):
+            fin = pos + n * 4
+            return (fin, *struct.unpack(end + 'I' * n, desc[pos:fin]))
 
-        pos += arch_bytes * 2
-        usec = struct.unpack("<" + fmt, desc[pos:pos+arch_bytes])[0] * 1000
-        result['pr_cutime_usec'] = struct.unpack("<" + fmt, desc[pos+arch_bytes:pos+arch_bytes*2])[0] + usec
+        def read_timeval():
+            sec, usec = struct.unpack(end+fmt*2, desc[pos:pos+2*arch_bytes])
+            return (pos+2*arch_bytes, sec * 1000000 + usec)
 
-        pos += arch_bytes * 2
-        usec = struct.unpack("<" + fmt, desc[pos:pos+arch_bytes])[0] * 1000
-        result['pr_cstime_usec'] = struct.unpack("<" + fmt, desc[pos+arch_bytes:pos+arch_bytes*2])[0] + usec
+        result = {}
 
-        pos += arch_bytes * 2
+        pos, result['si_signo'], result['si_code'], result['si_errno'] = read_ints(3)
+
+        # this field is a short, but it's padded to an int
+        result['pr_cursig'], = struct.unpack(end + "H", desc[pos:pos+2])
+        pos += 4
+
+        pos, result['pr_sigpend'], result['pr_sighold'] = read_longs(2)
+
+        pos, result['pr_pid'], result['pr_ppid'], result['pr_pgrp'], result['pr_sid'] = read_ints(4)
+
+        pos, result['pr_utime_usec'] = read_timeval()
+        pos, result['pr_stime_usec'] = read_timeval()
+        pos, result['pr_cutime_usec'] = read_timeval()
+        pos, result['pr_cstime_usec'] = read_timeval()
 
         # parse out general purpose registers
         if self.arch.name == 'AMD64':
@@ -179,34 +238,26 @@ class ELFCore(ELF):
         else:
             raise CLECompatibilityError("Architecture '%s' unsupported by ELFCore" % self.arch.name)
 
-        regvals = []
-        for idx in range(pos, pos+nreg*arch_bytes, arch_bytes):
-            regvals.append(struct.unpack("<" + fmt, desc[idx:idx+arch_bytes])[0])
+        assert nreg == len(rnames), "Please create an issue with this core-file attached to get this fixed."
+        pos, *regvals = read_longs(nreg)
         result['registers'] = dict(zip(rnames, regvals))
         del result['registers']['xxx']
 
-        pos += nreg * arch_bytes
-        result['pr_fpvalid'] = struct.unpack("<I", desc[pos:pos+4])[0]
+        pos, result['pr_fpvalid'] = read_ints(1)
+        assert pos <= len(desc) < pos + arch_bytes, "Please create an issue with this core-file attached to get this fixed."
+
         self.__current_thread.update(result)
 
+    def __parse_prpsinfo(self, desc):
+        pr_fname = desc.pr_fname.split(b'\x00', 1)[0]
+        try:
+            self.pr_fname = pr_fname.decode()
+        except UnicodeDecodeError:
+            self.pr_fname = repr(pr_fname)
+
     def __parse_files(self, desc):
-        self.filename_lookup = [(ent.vm_start, ent.vm_end, ent.page_offset * desc.page_size, fn.decode()) for ent, fn in zip(desc.Elf_Nt_File_Entry, desc.filename)]
-
-        # TODO this can be less stupid if we just parse out what the name/address of the main executable is
-        # that metadata has to be somewhere, right?
-        matched = None
-        if self.filename_lookup and self._main_filepath is not None:
-            for i, (a, b, c, fn) in enumerate(self.filename_lookup):
-                if os.path.basename(self._main_filepath) == fn[fn.rfind('/')+1:]: # explicit unix basename
-                    matched = fn
-                    break
-            else:
-                matched = self.filename_lookup[0][-1]
-
-        for i, (a, b, c, fn) in enumerate(self.filename_lookup):
-            if fn == matched:
-                self.filename_lookup[i] = (a, b, c, self._main_filepath)
-
+        self._page_size = desc.page_size
+        self.filename_lookup = [(ent.vm_start, ent.vm_end, ent.page_offset * desc.page_size, self._remote_file_mapper(fn.decode())) for ent, fn in zip(desc.Elf_Nt_File_Entry, desc.filename)]
 
     def __parse_x86_tls(self, desc):
         self.__current_thread['segments'] = {}
@@ -236,8 +287,7 @@ class ELFCore(ELF):
             self.auxv[code_str] = value
 
     def __reload_children(self):
-        # god damn. hacks start here
-        self.loader.page_size = 0x1000
+        self.loader.page_size = self._page_size
         self.loader._perform_relocations = False
 
         # hack: we are using a loader internal method in a non-kosher way which will cause our children to be
@@ -261,8 +311,14 @@ class ELFCore(ELF):
             try:
                 with open(filename, 'rb') as fp:
                     obj = self.loader._load_object_isolated(fp)
-            except FileNotFoundError:
-                l.warning("Could not load %s; core may be incomplete", filename)
+            except (FileNotFoundError, CLECompatibilityError) as ex:
+                if isinstance(ex, FileNotFoundError):
+                    l.warning("Dependency %s does not exist on the current system; this core may be incomplete.",
+                              filename)
+                elif isinstance(ex, CLECompatibilityError):
+                    l.warning("Could not find a compatible loader for %s; this core may be incomplete.", filename)
+                else:
+                    l.warning("Could not load %s; this core may be incomplete.", filename)
                 if self.loader.main_object is self:
                     self.loader.main_object = None
                 self.child_objects.clear()
@@ -307,38 +363,64 @@ class ELFCore(ELF):
                 self.child_objects.clear()
                 return
 
-            # store data provided by core into object
-            for vaddr, _, patch in patches:
-                try:
-                    obj.memory.store(vaddr - base_addr, patch)
-                except KeyError:
-                    pass  # this case handled below in the inject clause, right???
-
             obj._custom_base_addr = base_addr
             self.child_objects.append(obj)
 
-            # remove any core segments which are handled by this object
-            for seg in obj.segments:
-                addr = AT.from_lva(seg.vaddr, obj).to_rva() + base_addr
-                for subaddr in range(addr, addr + seg.memsize, 0x1000):
-                    match_seg = self.find_segment_containing(subaddr)
-                    if match_seg is not None:
-                        try:
-                            remaining_segments.remove(match_seg)
-                        except ValueError:
-                            pass
+            # figure out how the core's data should affect the child object's data
+            # iterate over all the core segments, since the only time we will need to make a change to the child's memory is if the core has something to say about it
+            # if there is ANY OVERLAP AT ALL, copy over the relevant data and nuke the segment
+            # then, if there is any part of the segment which DOESN'T correspond to a child segment, inject a new memory backer into the child for the relevant data
 
-            # inject any core segments which are not handled by the object but overlap with it
             max_addr = base_addr + (obj.max_addr - obj.min_addr)
             i = 0
             while i < len(remaining_segments):
                 seg = remaining_segments[i]
+                # check for overlap (overapproximation)
                 if base_addr <= seg.vaddr <= max_addr or seg.vaddr <= base_addr < seg.vaddr + seg.memsize:
                     remaining_segments.pop(i)
 
-                    seg_vaddr, backer = next(self.memory.backers(AT.from_mva(seg.vaddr, self).to_rva()))
-                    assert seg_vaddr == AT.from_mva(seg.vaddr, self).to_rva()
-                    obj.memory.add_backer(seg.vaddr - base_addr, backer)
+                    # if there is data before the beginning of the child or after the end, make new artificial segments for it
+                    if seg.vaddr < base_addr:
+                        size = base_addr - seg.vaddr
+                        remaining_segments.insert(i, Segment(seg.offset, seg.vaddr, size, size))
+                        i += 1
+                    if seg.max_addr > max_addr:
+                        size = seg.max_addr - max_addr
+                        offset = seg.memsize - size
+                        remaining_segments.insert(i, Segment(seg.offset + offset, seg.vaddr + offset, size, size))
+                        i += 1
+
+                    # ohhhh this is SUCH a confusing address space-conversation problem!
+                    # we're going to enumerate the contents of the core segment. at each point we find the relevant child backer. if this skips any content, inject a backer into the child.
+                    # then, copy the contents of the core segment that overlaps the child backer.
+                    cursor = max(0, base_addr - seg.vaddr)
+                    while cursor < seg.filesize:  # use filesize and not memsize so we don't overwrite stuff with zeroes if it's omitted from the core
+                        child_cursor = cursor + seg.vaddr - base_addr
+                        try:
+                            child_offset, child_backer = next(obj.memory.backers(child_cursor))
+                        except StopIteration:
+                            # is this right? is there any behavior we need to account for in the case that there is somehow no backer past a point mapped by the core?
+                            break
+
+                        # have we skipped any part of the core?
+                        skip_size = child_offset - child_cursor
+                        if skip_size > 0:
+                            # inject it into the child
+                            obj.memory.add_backer(child_cursor, self.memory.load(AT.from_mva(cursor + seg.vaddr, self).to_rva(), skip_size))
+
+
+                        # how much of the child's segment have we skipped by starting at the beginning of the core segment?
+                        child_backer_offset = max(0, -skip_size)
+                        # how much of the core's segment have we skipped and handled via injection?
+                        core_backer_offset = max(0, skip_size)
+                        # how much can we copy?
+                        copy_size = min(len(child_backer) - child_backer_offset, seg.memsize - (cursor + core_backer_offset))
+                        if copy_size > 0:
+                            # do the copy if we have anything to copy
+                            obj.memory.store(child_offset + child_backer_offset, self.memory.load(AT.from_mva(seg.vaddr + cursor + core_backer_offset, self).to_rva(), copy_size))
+
+                        # advance cursor
+                        cursor += core_backer_offset + copy_size
                 else:
                     i += 1
 
@@ -355,7 +437,22 @@ class ELFCore(ELF):
         self.has_memory = False
         if self.loader.main_object is self:
             self.loader.main_object = None
+            self.__record_main_object()
 
+    def __record_main_object(self):
+        """
+        If children objects are reloaded, identify the main object for later use by loader
+        """
+        for obj in self.child_objects:
+            if self.pr_fname and obj.binary_basename.startswith(self.pr_fname):
+                self._main_object = obj
+                return
+            if self._main_filepath is not None and os.path.basename(self._main_filepath) == obj.binary_basename:
+                self._main_object = obj
+                return
+
+        l.warning("Failed to identify main object in ELFCore")
+        self._main_object = self
 
 
 auxv_codes = {
